#Nathan Mellor 2016

import sys
from numpy import array,append,shape
from numpy.linalg import norm
from numpy import isnan
import lxml.etree as etree


def makeTemplate(infile):
    def parseElement(elm,td):
        if type(elm.tag)==str:
            if 'fileName' in elm.attrib:
                reldir='/'.join(infile.split('/')[:-1])
                if len(reldir)>0:
                    reldir+='/'
                xtree=etree.parse('%s%s' %(reldir,elm.attrib['fileName']))
                xroot=xtree.getroot()
                for child in xroot:
                    if type(child.tag)==str:
                        td=parseElement(child,td)
            else:
                if 'name' in elm.attrib:
                    base=elm.attrib['name']
                elif 'path' in elm.attrib:
                    base=elm.attrib['path'].split('/')[-1]
                elif 'name_colum2' in elm.attrib:
                    base=elm.attrib['name_colum2']

                if base not in td:
                    td[base]={}
                    if elm.text!=None:
                        eltxt=elm.text.strip()
                        if len(eltxt)==0:
                            eltxt=None
                    else:
                        eltxt=None
                    td[base]['self']=(elm.tag,elm.attrib,eltxt)
                for child in elm:
                    td[base]=parseElement(child,td[base])

        return td


    full_template={}
    tree=etree.parse(infile)
    root=tree.getroot()
    for node in root:
        full_template=parseElement(node,full_template)

    return full_template


def pp2xml(pp,outfile,inputTemplate,plantAge):
    def printLine(td,level):

        selfStr=td['self']

        iTabs=''
        for i in range(level):
            iTabs+='\t'
        strSt = '%s<%s ' % (iTabs,selfStr[0])

        pairs=[]
        for aname,aval in selfStr[1].items():
            pair='%s="%s"' % (aname,aval)
            pairs.append(pair)

        keyNames=['growthpoint','branches','dataPoints','rootLongitudinalGrowth','rootType']
        if 'name' in selfStr[1]:
            bName=selfStr[1]['name']
        else:
            bName='unknown'
        if len(td)==1 and bName not in keyNames and selfStr[2]==None:
            strEnd='/>\n'
        else:
            strEnd='>\n'
        sr_file.write(strSt+' '.join(pairs)+strEnd)
        level+=1

        if bName == 'growthpoint':
            sr_file.write('0 0 0 0 0 1 0\n5 0 0.5 0 0 1 0\n10 0 1 0 0 1 0\n')

        if bName == 'rootLongitudinalGrowth' and selfStr[1]['function']=='scaledRootGrowthRate':
            sr_file.write('0 0 0.1\n5 0.5 0.1\n10 1 0.1')

        if bName == 'rootType':
            sr_file.write('hypocotyl\n')

        if selfStr[2]!=None and bName not in keyNames:
            sr_file.write(selfStr[2]+'\n')

        for baseName,baseVals in td.items():
            if baseName!='self':
                printLine(baseVals,level)
        if len(td)>1 or bName in keyNames or selfStr[2]!=None:
            sr_file.write('%s</%s>\n' % (iTabs,selfStr[0]))

    def printHypocotyl(td,level,xid,name,degree,nodal_root_count,pp,selfval):

        keyNames=['growthpoint','branches','dataPoints','rootLongitudinalGrowth','rootType','rootDiameter','rootRSMLId']
        type_names=['tillerSeminal','tillerCrown']


        selfStr=td['self']
        if 'name' in selfStr[1]:
            bName=selfStr[1]['name']
        else:
            bName='unknown'


        iTabs=''
        for i in range(level):
            iTabs+='\t'
        strSt = '%s<%s ' % (iTabs,selfStr[0])

        pairs=[]
        for aname,aval in selfStr[1].items():
            if bName=='hypocotylTemplate' and aname=='name':
                pair='%s="%s"' % (aname,name)
            else:
                pair='%s="%s"' % (aname,aval)
            pairs.append(pair)

        tagOpen=True
        if len(td)==1 and bName not in keyNames:
            strEnd='/>\n'
            tagOpen=False
        elif bName=='branches':
            strEnd=' objectGenerator="insertRootBranchContainers">\n'
        elif bName=='rootDiameter':
            strSt = '%s<%s ' % (iTabs,'SimulaTable')
            selfStr=('SimulaTable',selfStr[1])
            strEnd=' name_column1="time" unit_column1="day" unit_column2="cm">\n'
        else:
            strEnd='>\n'

        sr_file.write(strSt+' '.join(pairs)+strEnd)
        level+=1
    
        branchTemplate=full_template['hypocotylTemplate'].copy()

        branchTemplate['self']=('SimulaConstant', {'name': 'hypocotylTemplate','type':'Coordinate','startTime':str(startTimes[xid])})

        if selfval is not None:
            sr_file.write(selfval)

        if bName =='rootType':
            sr_file.write('%s\n' % pp.rootTypes[xid])
        
        if bName == 'growthpoint':
            for tup7 in gDirections[xid]:   
                assert len(tup7)==7     
                sr_file.write(' '.join([str(x) for x in tup7])+'\n')

        if bName == 'rootLongitudinalGrowth':
            for tup3 in gRates[xid]:
                assert len(tup3)==3     
                sr_file.write(' '.join([str(x) for x in tup3])+'\n')

        if bName == 'rootDiameter':
            for tup3 in gRates[xid]:
                sr_file.write(str(tup3[0])+' '+str(pp.meanWidth[xid])+'\n')
            #diamlist=[str(0.0),str(pp.meanWidth[xid]),str(plantAge),str(pp.meanWidth[xid])]
            #sr_file.write(' '.join(diamlist)+'\n')
            #TODO: if/when root point widths are accurate, can use following lines instead of preceeding two
            #for tup2 in diameters[xid]:
            #   assert len(tup2)==2
            #   sr_file.write(' '.join([str(x) for x in tup2])+'\n')
    
        if bName == 'rootRSMLId':
            sr_file.write(str(xid)+'\n')            
        
        if bName == 'branches':
            nodal_root_count+=1
            if pp.rootTypes[xid]=='hypocotyl':
                for tillerTypeName,tillerIds in tillers.items():
                
                    sr_file.write('\t%s<SimulaBase name="%s" objectGenerator="rootBranches">\n' % (iTabs,tillerTypeName))
                    count=1
                    for rid in tillerIds:                   
                        branchTemplate['self'][1]['startTime']=str(startTimes[rid])
                        selfval=' '.join([str(x) for x in startPoints[rid]])+'\n'       
                        nodal_root_count = printHypocotyl(branchTemplate,level,rid,'%s_%s' %(tillerTypeName,count),degree,nodal_root_count,pp,selfval)
                        count+=1    
                    sr_file.write('\t%s</SimulaBase>\n' %iTabs)


        for baseName,baseVals in td.items():
            if baseName != 'self':
                nodal_root_count = printHypocotyl(baseVals,level,xid,name,degree,nodal_root_count,pp,None)
                
        if tagOpen==True:
            sr_file.write('%s</%s>\n' % (iTabs,selfStr[0]))
    
        return nodal_root_count




    def printRoot(td,level,xid,name,degree,root_count,pp,selfval):

        keyNames=['growthpoint','branches','dataPoints','rootLongitudinalGrowth','rootType','rootDiameter','rootRSMLId']
        type_names=['seminal','lateral','finelateral','finelateral2']


        selfStr=td['self']
        if 'name' in selfStr[1]:
            bName=selfStr[1]['name']
        else:
            bName='unknown'


        iTabs=''
        for i in range(level):
            iTabs+='\t'
        strSt = '%s<%s ' % (iTabs,selfStr[0])

        pairs=[]
        for aname,aval in selfStr[1].items():
            if bName=='siblingRootTemplate' and aname=='name':
                pair='%s="%s"' % (aname,name)
            else:
                pair='%s="%s"' % (aname,aval)
            pairs.append(pair)

        tagOpen=True
        if len(td)==1 and bName not in keyNames:
            strEnd='/>\n'
            tagOpen=False
        elif bName=='branches':
            strEnd=' objectGenerator="insertRootBranchContainers">\n'
        elif bName=='rootDiameter':
            strSt = '%s<%s ' % (iTabs,'SimulaTable')
            selfStr=('SimulaTable',selfStr[1])
            strEnd=' name_column1="time" unit_column1="day" unit_column2="cm">\n'
        else:
            strEnd='>\n'

        sr_file.write(strSt+' '.join(pairs)+strEnd)
        level+=1
    
        branchTemplate=full_template['siblingRootTemplate'].copy()

        branchTemplate['self']=('SimulaConstant', {'name': 'siblingRootTemplate','type':'Coordinate','startTime':str(startTimes[xid])})

        if selfval is not None:
            sr_file.write(selfval)

        if bName =='rootType':
            sr_file.write('%s\n' % pp.rootTypes[xid])
        
        if bName == 'growthpoint':
            for tup7 in gDirections[xid]:   
                assert len(tup7)==7     
                sr_file.write(' '.join([str(x) for x in tup7])+'\n')

        if bName == 'rootLongitudinalGrowth':
            for tup3 in gRates[xid]:
                assert len(tup3)==3     
                sr_file.write(' '.join([str(x) for x in tup3])+'\n')

        if bName == 'rootDiameter':
            for tup3 in gRates[xid]:
                sr_file.write(str(tup3[0])+' '+str(pp.meanWidth[xid])+'\n')
            #diamlist=[str(0.0),str(pp.meanWidth[xid]),str(plantAge),str(pp.meanWidth[xid])]
            #sr_file.write(' '.join(diamlist)+'\n')
            #TODO: if/when root point widths are accurate, can use following lines instead of preceeding two
            #for tup2 in diameters[xid]:
            #   assert len(tup2)==2
            #   sr_file.write(' '.join([str(x) for x in tup2])+'\n')
        
        if bName == 'rootRSMLId':
            sr_file.write(str(xid)+'\n')            
        
        if bName == 'branches':
            root_count+=1
            
            #************************************************
            if pp.rootTypes[xid]=='primaryRoot':
                if len(seminals)>0:
                    sr_file.write('\t%s<SimulaBase name="seminal" objectGenerator="rootBranches">\n' %iTabs)
                    level+=1
                    count=1
                    for rid in pp.children[xid]:
                        if pp.rootTypes[rid]=='seminal':
                            branchTemplate['self'][1]['startTime']=str(startTimes[rid])
                            selfval=' '.join([str(x) for x in startPoints[rid]])+'\n'
                            root_count = printRoot(branchTemplate,level,rid,'seminal_%s' % count,degree,root_count,pp,selfval)
                            count+=1
                    sr_file.write('\t%s</SimulaBase>\n' %iTabs)


            ordered_children=[]
            for rid in pp.children[xid]:
                if rid not in seminals:
                    ordered_children.append(rid)


            if len(ordered_children)>0:
                cdegree=degree+1
                sr_file.write('\t%s<SimulaBase name="%s" objectGenerator="rootBranches">\n' %(iTabs,type_names[cdegree]))
                level+=1
                count=1
                for rid in ordered_children:                    
                    branchTemplate['self'][1]['startTime']=str(startTimes[rid])
                    selfval=' '.join([str(x) for x in startPoints[rid]])+'\n'       
                    root_count = printRoot(branchTemplate,level,rid,'%s_%s' %(type_names[cdegree],count),cdegree,root_count,pp,selfval)
                    count+=1    
                sr_file.write('\t%s</SimulaBase>\n' %iTabs)


        for baseName,baseVals in td.items():
            if baseName != 'self':
                root_count = printRoot(baseVals,level,xid,name,degree,root_count,pp,None)
                
        if tagOpen==True:
            sr_file.write('%s</%s>\n' % (iTabs,selfStr[0]))
    
        return root_count






    #START OF MAIN ROUTINE*******************************

    #1. DO SOME CHECKS AND TIDY UP IF NEEDED

    #identify primary and seminals (if present)
    seminals=[]
    for rid,rtype in pp.rootTypes.items():
        if rtype=='primaryRoot' or rtype=='primary':
            prim_id=rid
        if rtype=='seminal':
            seminals.append(rid)
        if rtype=='hypocotyl':
            hypocotyl_id=rid

    if pp.rootTypes[prim_id]=='primary':
        pp.rootTypes[prim_id]='primaryRoot'


    #identify tillers (if present)
    tillers={}
    for rid,rtype in pp.rootTypes.items():
        if rtype[:6]=='tiller':
            if rtype not in tillers:
                tillers[rtype]=[rid]
            else:
                tillers[rtype].append(rid)

    #translate all the points simRoot has -y with depth not +z
    for rid,pts in pp.rootPoints.items():
        newpts=[]
        for pt in pts:
            newpts.append([pt[0],-pt[2],pt[1]])
        pp.rootPoints[rid]=newpts
    
    strip=5 #TODO hardcoded
    #strip out some points (if strip >1)
    if strip>1:
        for rid,pts in pp.rootPoints.items():
            newpts=[]
            branchPoints=[pp.rootPoints[xid][0] for xid in pp.children[rid]]#TODO: could run into problems here if strip>1 and tillering roots have children
            for i,pt in enumerate(pts):
                if i%strip==0 or pt in branchPoints:
                    newpts.append(pt)
            if (len(pts)-1)%strip!=0:
                newpts.append(pts[-1])
            pp.rootPoints[rid]=newpts
            if len(newpts)<2:
                print('len < 2',len(pts))
            #print 'max l',max([norm(array(newpts[i-1])-array(newpts[i])) for i in range(1,len(newpts))])

    #recalculate length (if strip >1)
    for rid,pts in pp.rootPoints.items():
        rlength=0.0
        for i in range(1,len(pts)):
            rlength+=norm(array(pts[i])-array(pts[i-1]))
        pp.length[rid]=rlength

    #check widths are not greater than parents
    for rid,wid in pp.meanWidth.items():
        try:
            if wid>pp.meanWidth[pp.parent[rid]]:
                pp.meanWidth[rid]=pp.meanWidth[pp.parent[rid]]  
        except KeyError:#no parent
            pass    

    #check branches only have one parent (TODO: is this still needed?)
    branched=[]
    for rid,kids in pp.children.items():
        for kid in kids:
            if kid in branched:
                print('problem',kid)
            branched.append(kid)

    #2. FIND START TIMES

    lengthAtTime=pp.lengthAtTime
    times=[]
    timedict={}
    for rid,timelens in lengthAtTime.items():
        for time in timelens.keys():
            if time not in times:
                timedict[time]=[rid]
                times.append(time)
            else:
                timedict[time].append(rid)
    times=sorted(times)
    plantAge=max(times)
    print('**********************',plantAge)
    #print times
    #print(timedict)
    #print(pp.rootTypes)
    #print tillers
    #print pp.rootTypes
    startTimes={hypocotyl_id:0.0}
    lasttime=0.0
    count=0
    pp.rootTypes[prim_id]='seminal'
    for time in times:
        #print 'time=',time
        for ttype in ['seminal','lateral']:
            #print(ttype,time)
            
            newlengths=[]
            for rid in timedict[time]:
                if rid not in startTimes and pp.rootTypes[rid]==ttype:
                    newlengths.append(lengthAtTime[rid][time])
            if len(newlengths)>0:               
                maxLengthAtTime=max(newlengths)
            else:
                continue
            print(maxLengthAtTime)
            currentRate=maxLengthAtTime/float(time-lasttime)
            #print 'crate',currentRate
            for rid in timedict[time]:
                if rid not in startTimes and pp.rootTypes[rid]==ttype:
                    count+=1
                    startTimes[rid]=time-lengthAtTime[rid][time]/currentRate
                    #print rid,startTimes[rid]
        lasttime=time
            #print 'count',count
    pp.rootTypes[prim_id]='primaryRoot'
    startTimes[prim_id]=0.0     
        

        
    #print(startTimes)

    

    #3. SET START POINTS
    startPoints={}

    #set start point to be relative to start point of parent
    def transPoints(xid,points,shift):
        points[xid]=list(array(points[xid])-array(shift))
        newshift=points[xid]
        for qid in pp.children[xid]:
            points[qid]=list(array(points[qid])-array(shift))
            points=transPoints(qid,points,newshift)
        return points

    #take first point as start point initially
    startPoints={rid:pts[0] for rid,pts in pp.rootPoints.items()}

    #translate so relative to parent start
    startPoints=transPoints(prim_id,startPoints,pp.rootPoints[prim_id][0])



    #translate points so relative to start point
    for rid,pts in pp.rootPoints.items():
        newpts=[]
        for pt in pts:
            newpts.append(list(array(pt)-array(pts[0])))
        pp.rootPoints[rid]=newpts



    
    #4. GET GROWTHPOINTS, RATES and DIAMETERS
    #set up empty dictionaries to store 'growthpoint', 'rootLongitudinalGrowth',and 'rootDiameter' data
    gDirections={}
    gRates={}   
    diameters={}



    for rid,pts in pp.rootPoints.items():

        #empty lists for given root id
        gDirections[rid]=[]
        gRates[rid]=[]
        diameters[rid]=[]

        lat=lengthAtTime[rid]

        currentLength=0.0
        lastTimepoint=startTimes[rid]
        newTimepoint=float(min(lat.keys()))
        thisTimepoints=sorted(lat.keys())
        lenChange=lat[newTimepoint]
        lastLAT=0
        timeref=0
        rate=lenChange/(newTimepoint-lastTimepoint)
        
        #first to penultimate point
        for z,pt in enumerate(pts[:-1]):

            #direct to next point
            direct=array(pts[z+1])-array(pt)                
            #time of current point
            currentTime=lastTimepoint+(currentLength-lastLAT)/rate
            #append to growthrates
            gRates[rid].append([currentTime,currentLength,rate])
            gDirections[rid].append([currentTime,pt[0],pt[1],pt[2],direct[0],direct[1],direct[2]])
            diameters[rid].append([currentTime,pp.pointWidth[rid][z]])

            #increase current length
            currentLength+=norm(direct)
            
            if currentLength-lat[newTimepoint]>0.001:
                lastTimepoint=float(thisTimepoints[timeref])
                newTimepoint=float(thisTimepoints[timeref+1])
                lenChange=lat[newTimepoint]-lat[lastTimepoint]
                lastLAT=lat[lastTimepoint]
                timeref+=1
                rate=lenChange/(newTimepoint-lastTimepoint)
                
        #final point
        #time of current point
        gDirections[rid].append([max(lat.keys()),pts[-1][0],pts[-1][1],pts[-1][2],direct[0],direct[1],direct[2]])
        gRates[rid].append([max(lat.keys()),currentLength,rate])
        diameters[rid].append([max(lat.keys()),pp.pointWidth[rid][-1]])
                


    #5. MAKE THE FILE


    #get xml template dict
    full_template=makeTemplate(inputTemplate)
    full_template['shootTemplate']['self'][1]['name']='shoot'
    full_template['hypocotylTemplate']['self'][1]['name']='hypocotyl'
    full_template['hypocotylTemplate']['rootType']={'self': ('SimulaConstant', {'type': 'string','name': 'rootType'},None)}
    full_template['siblingRootTemplate']['rootType']={'self': ('SimulaConstant', {'type': 'string','name': 'rootType'},None)}


    #open output file with correct extension
    if outfile[-4:]=='.xml':
        sr_file=open(outfile,"w")
    else:
        sr_file=open('%s.xml' % outfile,"w")

    #write header
    sr_file.write('<SimulationModelIncludeFile>\n')

    #write parent SimulaBase for whole plant
    systemName='simpleRoot' #TODO: hardcoded
    sr_file.write('\t<SimulaBase name="%s">\n' % systemName)

    #write plant type
    sr_file.write('\t\t<SimulaConstant name="plantType" type="string">\n')
    sr_file.write('\t\t\t%s\n' % systemName) #TODO: plant type could be different from systemName
    sr_file.write('\t\t</SimulaConstant>\n')

    #write planting time
    plantingTime=0 #TODO: hardcoded
    sr_file.write('\t\t<SimulaConstant name="plantingTime" unit="day" type="Time">\n')
    sr_file.write('\t\t\t%d\n' % plantingTime)
    sr_file.write('\t\t</SimulaConstant>\n')

    #write plant position (everything else hangs off this)
    plantPos=(0,-2,0) #TODO: hardcoded
    sr_file.write('\t\t<SimulaConstant name="plantPosition" type="Coordinate">\n')
    sr_file.write('\t\t\t%.1f %.1f %.1f\n' % plantPos)

    #write shoot
    printLine(full_template['shootTemplate'],3)

    #write hypocotyl + tillers
    nodal_root_count=printHypocotyl(full_template['hypocotylTemplate'],3,hypocotyl_id,'hypocotyl',0,0,pp,None)

    #write root system starting from primary
    root_count=printRoot(full_template['siblingRootTemplate'],3,prim_id,'primaryRoot',0,0,pp,None)
    print('primary borne roots',root_count)
    print('tillers',nodal_root_count)
    print('roots in pp',len(pp.rootPoints))


    #close plant position
    sr_file.write('\t\t</SimulaConstant>\n')

    #write whole plant derivatives
    for baseName,baseVals in full_template['plantTemplate'].items():
        if baseName!='self':
            printLine(baseVals,2)
    #close whole plant
    sr_file.write('\t</SimulaBase>\n')

    #write footer
    sr_file.write('</SimulationModelIncludeFile>\n')
    
    sr_file.close()
    








