    #Nathan Mellor 2016

#import rsml
#from rsml.measurements import *
import lxml.etree as etree
import sys,os,socket,datetime,time
from pyPlant import pyPlant
from numpy import array
from scipy.linalg import norm
import csv

from scipy.cluster.vq import kmeans,vq

def eMess(errstr):
    print('EXITING:',errstr)
    sys.exit()

def rsml2pp(filename):

    #open file as etree
    intree=etree.parse(filename)
    #print(filename)
    
    
    #find and set metadata
    metadata=intree.find('metadata')
    if metadata==None:
        print('no metadata found in input file')
        sys.exit()
    else:
        unit = metadata.find('unit').text
        resolution = metadata.find('resolution').text
    #print('metadata found')
    if unit=='cm':
        scale_to_cm=1.0
    elif unit=="pixel":
        #print('*using scale of 54um per pixel*')
        scale_to_cm=0.0054#TODO hardcoded - check from metadata somehow

    user = '%s@%s' % (os.environ.get('LOGNAME'),socket.gethostname())#user name from system (switch with a hard-coded string if problems)
    ts = datetime.datetime.fromtimestamp(time.time()).strftime('%d-%m-%Y_T%H:%M:%S')#current date and time


    #TODO: units are hardcoded
    mDict={'version':'0','unit':'cm','resolution':str(float(resolution)*scale_to_cm),'last-modified':ts,'software':sys.argv[0],'user':user,'file-key':filename.split('/')[-1]}
    #print('metadata:',mDict)


    #check scene 
    scene=intree.find('scene')
    if scene==None:
        eMess('no scene in input file')
    for plant in scene:
        assert plant.tag=='plant'

    #check one and only one plant       
    plants=[el for el in scene.iter('plant')]

    if  len(plants)>1:
        #TODO: deal with >1 plant
        eMess('more than one plant in input file')
    elif len(plants)==0:
        eMess('zero plants in input file')
    
    plant=plants[0]

    #check/read roots
    children={}
    rTypes={}
    typeLists={}
    meanWidths={}
    axAngles={}
    lengths={}
    rootPoints={}
    pointWidths={}


    #find primary
    prim_id=None
    for root in plant.iter('root'):

        rid=root.attrib['id']
        if root.attrib['label']=='primary':
            prim_id=int(rid)
    if prim_id==None:
        eMess('no primary root annotated')

    #print('primary root found. id is',prim_id)
    

    
    for root in plant.iter('root'):

        rid=root.attrib['id']

        # get root connectivity
        if int(rid) in children:
            eMess('non-unique root id: %s' % rid)

        children[int(rid)]=[]
        for child in root.iterchildren('root'):
            children[int(rid)].append(int(child.attrib['id']))

        # get properties
        properties=root.find('properties')

        
        #PROPERTY: type 
        rtype=properties.find('type')
        cType=rtype.attrib['value']

        rTypes[int(rid)]=cType
        if cType in typeLists:
            typeLists[cType].append(int(rid))
        else:
            typeLists[cType]=[int(rid)]
        
        #PROPERTY: width
        width=properties.find('width')
        try:
            meanWidths[int(rid)]=float(width.attrib['value'])*scale_to_cm
        except ValueError:
            #print(rid,'missing width data')
            meanWidths[int(rid)]=0.0
        
        #PROPERTY: angle
        angle=properties.find('angle')
        if angle==None:
            eMess('no angle property for root %s' % rid)
        axAngles[int(rid)]=float(angle.attrib['value'])
        
        #PROPERTY: length
        length=properties.find('length')
        if length==None:
            eMess('no length property for root %s' % rid)
        lengths[int(rid)]=float(length.attrib['value'])*scale_to_cm
        
        #get geometry
        geometry=root.find('geometry')
        if geometry==None:
            eMess('no geometry for root %s' % rid)
        polyline=geometry.find('polyline')
        if polyline==None:
            eMess('no polyline for root %s' % rid)
        rootPoints[int(rid)]=[]
        for pt in polyline.iterchildren('point'):
            rootPoints[int(rid)].append([float(pt.attrib['x']),float(pt.attrib['y']),float(pt.attrib['z'])])
        
        
    
        pointWidths[int(rid)]=[]
        
        #get diameters.
        functions=root.find('functions')
        
        
        if functions is not None:
            last=meanWidths[int(rid)]
            function=functions.find('function')
            for pt in function.iterchildren('sample'):
                try:
                    last=float(pt.attrib['value'])*scale_to_cm
                    pointWidths[int(rid)].append(float(pt.attrib['value'])*scale_to_cm)
                except ValueError:
                    pointWidths[int(rid)].append(last)
        else:
            #print('setting zero width')
            pointWidths[int(rid)]=[0.0 for pt in rootPoints[int(rid)]]
            
        #TODO: generalise properties/functions etc.

    
    #print(children)
    #print(typeLists)
    #print(lengths[2],lengths[3])
    
    #test primary
    if 'primary' not in typeLists and 'primaryRoot' not in typeLists:
        if prim_id!=None:
            for rtype,rids in typeLists.items():
                if prim_id in rids:
                    typeLists[rtype].remove(prim_id)
            typeLists['primary']=[prim_id]
            rTypes[prim_id]='primary'
        else:
            eMess('no primary root found')

    if 'primary' in typeLists:
        pstr='primary'
    else:
        pstr='primaryRoot'

    extraPrims=[]
    if len(typeLists[pstr])>1:
        print('EXTRA PRIMARY')
        testlist=[lengths[xid] for xid in typeLists[pstr]]
        prim_id=typeLists[pstr][testlist.index(max(testlist))]
        for xid in typeLists[pstr]:
            if xid != prim_id:
                extraPrims.append(xid)
        typeLists[pstr]=[prim_id]
    

    prim_id=typeLists[pstr][0]
    
    #translate and scale all points based on given scale and origin point of primary
    zero=array(rootPoints[prim_id][0])
    if len(rootPoints[prim_id][0])==2:
        zero=array([zero[0],0.0,zero[1]])

    for rid, vals in rootPoints.items():
        for i,val in enumerate(vals):
            nval=array(val)
            if len(val)==2:
                nval=array([nval[0],0.0,nval[1]])
            rootPoints[rid][i]=scale_to_cm*(nval-zero)
    
    #print(len(rootPoints),len(typeLists))
    
    #test seminals
    """
    if 'seminal' not in typeLists:
        typeLists['seminal']=[]
        sem_found=False

        for xid,xtype in csv_types.items():
            if xtype=='seminal':
                sem_found=True
                for rtype,rids in typeLists.items():
                    if xid in rids:
                        typeLists[rtype].remove(xid)
                typeLists['seminal'].append(xid)
                rTypes[xid]='seminal'
        #eMess('no seminal roots found')            
        if sem_found==False:
            trans={i:rid for i,rid in enumerate(csv_lengths.keys())}
            print(filename,'Warning: No seminals found - using k means clustering')
            centroids,_ = kmeans(array(csv_lengths.values()),2)

            highval = list(centroids).index(max(list(centroids)))
            idx,_ = vq(array(csv_lengths.values()),centroids)

            sems=[]
            for i,val in enumerate(list(idx)):
                if val==highval:
                    sems.append(trans[i])
            for sem in sems:
                if sem !=prim_id:
                    for rtype,rids in typeLists.items():
                        if sem in rids:
                            typeLists[rtype].remove(sem)
                    typeLists['seminal'].append(sem)
                    rTypes[sem]='seminal'
    """
    
       
    if 'seminal' in typeLists:
        for rid in typeLists['seminal']:
            if rid not in children[prim_id]:
                for xid,xc in children.items():
                    if rid in xc:
                        children[xid].remove(rid)
                
                children[prim_id].append(rid)
    for rtype in typeLists.keys():
        if rtype not in ['primary','seminal','lateral']:
            print(rtype)
    #check for zero point roots
    badRoots=[]
    for rid,pts in rootPoints.items():
        if len(pts)<2:
            badRoots.append(rid)
        #if lengths[rid]==0.0:
            #print('zero point root',filename,rid,pts)
    #if len(badRoots)>0:
        #print(badRoots)
    
    #check and rename higher order roots
    """
    def getOrder(rid,level,orders):
        orders[rid]=level
        level+=1
        for kid in children[rid]:
            if rTypes[kid]=='seminal':
                orders=getOrder(kid,level-1,orders)
            else:
                orders=getOrder(kid,level,orders)
        return orders
    try:
    orders=getOrder(prim_id,0,{})   
    for rid,order in orders.items():
        if order==2:
            rTypes[rid]='finelateral'
        if order==3:
            rTypes[rid]='finelateral2'
        if order>3:
            print(filename,'order',order)
    except:
        quit()

    """
    #remove bad roots
    for rid in badRoots:
        for xid,kids in children.items():
            if rid in kids:
                children[xid].remove(rid)
        del children[rid]
        del rTypes[rid]
        del lengths[rid]
        del axAngles[rid]
        del meanWidths[rid]
        del pointWidths[rid]
        del rootPoints[rid]

    
    def alternateParent(rid,rootPoints):
        #try other end of root
        testPt=rootPoints[rid][-1]
        #default origin of primary
        mindiff=norm(testPt-rootPoints[prim_id][0])
        aid=prim_id
        mini=0
        #test all other points
        for xid,pts in rootPoints.items():
            if xid!=rid:
                for i,pt in enumerate(pts):
                    diff=norm(testPt-pt)
                    if diff<mindiff:
                        mindiff=diff
                        aid=xid
                        mini=i+1
        print('alternate root:',aid, 'type:',rTypes[aid],'distance:',mindiff,'cm')
        rootPoints[rid].reverse()
        return aid,mindiff,mini

    #add start points if needed and order branches
    branchIds={}
    changedParent={}
    for rid,kids in children.items():
        
        kpos={}
        rpos={}
        branchIds[rid]=[]
        for kid in kids:
            ppts=rootPoints[rid]
            kpts=rootPoints[kid]
            mindiff=norm(kpts[0]-ppts[0])
            mini=0
            for i,pt in enumerate(ppts[1:]):
                diff=norm(kpts[0]-pt)
                if diff<mindiff:
                    mindiff=diff
                    mini=i


            #check root not attched to end of another
            #TODO: deal with this better somehow
            if mini==len(ppts)-2:
                print(filename,'root',kid,'attached to end of root',rid,':')
                aid,mindiff,mini=alternateParent(kid,rootPoints)
                ppts=rootPoints[aid]
                changedParent[kid]=(rid,aid,mini)

            else:
                if mini in kpos:
                    kpos[mini].append(kid)
                else:
                    kpos[mini]=[kid]
                rpos[kid]=mini


            if mindiff>0.0:
                rootPoints[kid].insert(0,ppts[mini+1])
        if len(kpos)>0:
            children[rid]=[]
            for ref in sorted(kpos):
                for kid in kpos[ref]:
                    if kid not in badRoots:
                        children[rid].append(kid)
                        branchIds[rid].append(rpos[kid])
    #move roots
    for rid,vals in changedParent.items():
        if rid not in badRoots:
            if len(branchIds[vals[1]])>0:
                i=0
                while branchIds[vals[1]][i]<vals[2]:
                    i+=1
                branchIds[vals[1]].insert(i,vals[2])
                children[vals[1]].insert(i,rid)
            else:
                branchIds[vals[1]].append(vals[2])
                children[vals[1]].append(rid)               
            if rid in children[vals[0]]:
                children[vals[0]].remove(rid)

    
    #add branch positions and length of branches            
    branchPositions={}
    branchLengths={}
    for rid,pts in rootPoints.items():
        branchPositions[rid]=[]
        branchLengths[rid]=[]
        plengths=[]
        cumlength=0.0
        for i,pt in enumerate(pts[1:]):
            cumlength+=norm(pt-pts[i])
            plengths.append(cumlength)
        lengths[rid]=cumlength#use measured length rather than imported
        for i,xid in enumerate(branchIds[rid]):
            branchPositions[rid].append(plengths[xid])
            branchLengths[rid].append(lengths[children[rid][i]])



            
    #TODO: check total length against imported length (plus branch lengths)
    for rid,pts in rootPoints.items():
        if lengths[rid]==0.0:
            print('zrp2',filename,rid,pts)
            

    #make pyPlant object. TODO: add functions to pyPlant that check the correct format / size object is given
    pp=pyPlant()
    pp.setConnectivity(children)
    pp.rootTypes=rTypes
    pp.length=lengths
    pp.branchPositions=branchPositions
    pp.branchLengths=branchLengths
    pp.axAngle=axAngles
    pp.meanWidth=meanWidths
    pp.pointWidth=pointWidths
    pp.rootPoints=rootPoints
    pp.metadata=mDict


    #test for order 1 laterals (i.e. laterals that have no parent root)
    for xid in children.keys():
        if xid not in pp.parent and xid != prim_id:
            eMess('WARNING: order 1 lateral found, %s' % xid)

    #add an arbitrary hypocotyl
    newID=-1
    pp.rootPoints[newID]=[[0.0,0.0,0.0],[0.0,0.0,-1.0],[0.0,0.0,-2.0]]
    pp.length[newID]=2.0
    pp.rootTypes[newID]='hypocotyl'
    pp.meanWidth[newID]=0.1 #TODO get estimate / data for this
    pp.pointWidth[newID]=[pp.meanWidth[newID] for pt in pp.rootPoints[newID]]
    pp.children[newID]=[]
    pp.branchPositions[newID]=[]
    pp.branchLengths[newID]=[]
    pp.axAngle[newID]=[]

    #return pyPlant object

    return pp



    
