## Griffiths et al 2021, X-ray CT reveals 4D root system development and lateral root responses to nitrate in soil

## RooTh traits output .csv to Summary statistics
## https://github.com/rootth/rooth (Johnson J)
## R code (York LM, Griffiths M)

#################
## User setup  ##
#################
# 1) Set working directory to folder where RooTh .csv's are
# 2) Check no output .csv already present
# 3) Check no other files in input folder except for RooTh .csv's
# 3) Install or load following packages

# Change output directory to input folder   #

library(tidyverse) #ggplot2, purrr, tibble, dplyr, tidyr, stringr, readr, forcats

#####################
## Create function ##
#####################
rooth2Rbyclass <- function(RTpath = "RTpath") {
  
  file_list<- list.files(path = RTpath, pattern=".csv") #make list of csv files in RTpath directory
  
  roothcolnames <- c("file.name"
                     ,"root.class"
                     ,"rth.length.mm.avg"
                     ,"rth.diam.um.avg"
                     ,"rth.volume.vox.avg"
                     ,"rth.angle.10.avg"
                     ,"rth.angle.25.avg"
                     ,"rth.angle.50.avg"
                     ,"rth.angle.100.avg"
                     ,"rth.number"
                     ,"rth.length.mm.total"
                     ,"rth.vol.vox.total"
  )
  
  for (file in file_list){ #go through all the files
    
    roothdata <- read.csv(paste(RTpath, file, sep="")) #read in file
    
    meanfactors <- with(roothdata, list(Root.Type)) #what to aggregate by
    roothsplittraits <- aggregate(roothdata[c(5,7:12)], by=meanfactors, FUN=mean) #get means within each root class
    roothsplittraits$number <- as.vector(table(roothdata$Root.Type)) #count the number of roots in each type
    roothsplittraits$length.total <- aggregate(roothdata$Root.Length.ACM..mm., by=meanfactors, FUN=sum)$x #calculate total length for each class
    roothsplittraits$vol.total <- aggregate(roothdata$Root.Volume..voxels., by=meanfactors, FUN=sum)$x #calculate total volume for each class
    roothsplittotal <- as.vector(apply(roothdata[c(5,7:12)], 2, mean)) #take means across all roots
    roothsplittotal <- append(roothsplittotal,dim(roothdata)[1]) #attach the number of roots of system
    roothsplittotal <- append(roothsplittotal,sum(roothdata[5])) #attach the total root length of system
    roothsplittotal <- append(roothsplittotal,sum(roothdata[7])) #attach teh total root volume of system
    roothsplittotal <- append(c("system"), roothsplittotal) #add 'system' name
    roothsplittraits$Group.1 <- as.character(roothsplittraits$Group.1 ) #assumes factors, need to change to character
    names(roothsplittraits)[1] <- "root.class" #change name to root.class
    roothsplittraitscombined <- rbind(roothsplittraits, roothsplittotal) #combined the within class traits with total system
    roothsplittraitscombined2 <- cbind(rep(file, dim(roothsplittraitscombined)[1]), roothsplittraitscombined) #add file names
    names(roothsplittraitscombined2) <- roothcolnames #change column names to what we want
    
    # if the merged dataset doesn't exist, create it
    if (!exists("dataset")){
      dataset <- roothsplittraitscombined2
    }
    
    # if the merged dataset does exist, append to it
    else {
      dataset<-as.data.frame(rbind(dataset, roothsplittraitscombined2))
    }
    
  } #end for loop
  
  row.names(dataset) <- 1:dim(dataset)[1] #give new rownames for size of dataset
  #dataset2 <- cbind(file_list, dataset) #attach the file names as the identity
  colnames(dataset) <- roothcolnames
  return(dataset)
} #end rooth2R function

##################
## Run function ##
##################
RooTh_trait_output <- rooth2Rbyclass(RTpath="C:/Users/USERNAME/FILEPATH/")  # PC choose folder containing RooTh output .csv's
RooTh_trait_output <- rooth2Rbyclass(RTpath="/Users/USERNAME/FILEPATH/")  # MacOS choose folder containing RooTh output .csv's

###################
## Create output ##
##   directory   ##
###################
dir.create("output",showWarnings = TRUE)

#################
## Save output ##
#################
name <- c("Rep1", "Day1", "Day2")

RooTh_trait_output <- RooTh_trait_output %>%
  separate(file.name
           ,into = c("ScanNumber", "Geno", "Treatment", "Rep", "Day")
           ,sep = '_') %>%
  separate(`Rep`
           ,into = c("Rep1","Rep")
           ,sep = 'R') %>%
  separate(`Day`
           ,into = c("Day1", "Day")
           ,sep = 'D') %>%
  separate(`Day`
           ,into = c("Day", "Day2")
           ,sep = '.csv') %>%
  select(-one_of(name))

write_csv(RooTh_trait_output, paste("output/RooTh_trait_output.csv"))